;;; This file is only used for installing AUC TeX.
;;; It is not a part of AUC TeX itself.

;; Make sure we get the right files.
(setq load-path (cons "." load-path)
      byte-compile-warnings nil
      TeX-lisp-directory "<none>")

